/**
 * Based on ndef.parser, by Raphael Graf <r@undefined.ch>
 * http://www.undefined.ch/mparser/index.html
 *
 * Ported to JavaScript and modified by Matthew Crumley <email@matthewcrumley.com>
 * https://silentmatt.com/javascript-expression-evaluator/
 *
 * Ported to QMLJS with modifications done accordingly done by Ad5001 <mail@ad5001.eu> (https://ad5001.eu)
 *
 * Copyright (c) 2015 Matthew Crumley, 2021-2024 Ad5001
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * You are free to use and modify this code in anyway you find useful. Please leave this comment in the code
 * to acknowledge its original source. If you feel like it, I enjoy hearing about projects that use my code,
 * but don't feel like you have to let me know or ask permission.
 */

export function add(a, b) {
    return Number(a) + Number(b)
}

export function sub(a, b) {
    return a - b
}

export function mul(a, b) {
    return a * b
}

export function div(a, b) {
    return a / b
}

export function mod(a, b) {
    return a % b
}

export function concat(a, b) {
    if(Array.isArray(a) && Array.isArray(b)) {
        return a.concat(b)
    }
    return "" + a + b
}

export function equal(a, b) {
    return a === b
}

export function notEqual(a, b) {
    return a !== b
}

export function greaterThan(a, b) {
    return a > b
}

export function lessThan(a, b) {
    return a < b
}

export function greaterThanEqual(a, b) {
    return a >= b
}

export function lessThanEqual(a, b) {
    return a <= b
}

export function andOperator(a, b) {
    return Boolean(a && b)
}

export function orOperator(a, b) {
    return Boolean(a || b)
}

export function inOperator(a, b) {
    return b.includes(a)
}

export function sinh(a) {
    return ((Math.exp(a) - Math.exp(-a)) / 2)
}

export function cosh(a) {
    return ((Math.exp(a) + Math.exp(-a)) / 2)
}

export function tanh(a) {
    if(a === Infinity) return 1
    if(a === -Infinity) return -1
    return (Math.exp(a) - Math.exp(-a)) / (Math.exp(a) + Math.exp(-a))
}

export function asinh(a) {
    if(a === -Infinity) return a
    return Math.log(a + Math.sqrt((a * a) + 1))
}

export function acosh(a) {
    return Math.log(a + Math.sqrt((a * a) - 1))
}

export function atanh(a) {
    return (Math.log((1 + a) / (1 - a)) / 2)
}

export function log10(a) {
    return Math.log(a) * Math.LOG10E
}

export function neg(a) {
    return -a
}

export function not(a) {
    return !a
}

export function trunc(a) {
    return a < 0 ? Math.ceil(a) : Math.floor(a)
}

export function random(a) {
    return Math.random() * (a || 1)
}

export function factorial(a) { // a!
    return gamma(a + 1)
}

export function isInteger(value) {
    return isFinite(value) && (value === Math.round(value))
}

const GAMMA_G = 4.7421875
const GAMMA_P = [
    0.99999999999999709182,
    57.156235665862923517, -59.597960355475491248,
    14.136097974741747174, -0.49191381609762019978,
    0.33994649984811888699e-4,
    0.46523628927048575665e-4, -0.98374475304879564677e-4,
    0.15808870322491248884e-3, -0.21026444172410488319e-3,
    0.21743961811521264320e-3, -0.16431810653676389022e-3,
    0.84418223983852743293e-4, -0.26190838401581408670e-4,
    0.36899182659531622704e-5
]

// Gamma function from math.js
export function gamma(n) {
    let t, x

    if(isInteger(n)) {
        if(n <= 0) {
            return isFinite(n) ? Infinity : NaN
        }

        if(n > 171) {
            return Infinity // Will overflow
        }

        let value = n - 2
        let res = n - 1
        while(value > 1) {
            res *= value
            value--
        }

        if(res === 0) {
            res = 1 // 0! is per definition 1
        }

        return res
    }

    if(n < 0.5) {
        return Math.PI / (Math.sin(Math.PI * n) * gamma(1 - n))
    }

    if(n >= 171.35) {
        return Infinity // will overflow
    }

    if(n > 85.0) { // Extended Stirling Approx
        const twoN = n * n
        const threeN = twoN * n
        const fourN = threeN * n
        const fiveN = fourN * n
        return Math.sqrt(2 * Math.PI / n) * Math.pow((n / Math.E), n) *
            (1 + (1 / (12 * n)) + (1 / (288 * twoN)) - (139 / (51840 * threeN)) -
                (571 / (2488320 * fourN)) + (163879 / (209018880 * fiveN)) +
                (5246819 / (75246796800 * fiveN * n)))
    }

    --n
    x = GAMMA_P[0]
    for(let i = 1; i < GAMMA_P.length; ++i) {
        x += GAMMA_P[i] / (n + i)
    }

    t = n + GAMMA_G + 0.5
    return Math.sqrt(2 * Math.PI) * Math.pow(t, n + 0.5) * Math.exp(-t) * x
}

export function stringOrArrayLength(s) {
    if(Array.isArray(s)) {
        return s.length
    }
    return String(s).length
}

export function hypot() {
    let sum = 0
    let larg = 0
    for(let i = 0; i < arguments.length; i++) {
        const arg = Math.abs(arguments[i])
        let div
        if(larg < arg) {
            div = larg / arg
            sum = (sum * div * div) + 1
            larg = arg
        } else if(arg > 0) {
            div = arg / larg
            sum += div * div
        } else {
            sum += arg
        }
    }
    return larg === Infinity ? Infinity : larg * Math.sqrt(sum)
}

export function condition(cond, yep, nope) {
    return cond ? yep : nope
}

/**
 * Decimal adjustment of a number.
 * From @escopecz.
 *
 * @param {number} value - The number.
 * @param {Integer} exp - The exponent (the 10 logarithm of the adjustment base).
 * @return {number} - The adjusted value.
 */
export function roundTo(value, exp) {
    // If the exp is undefined or zero...
    if(typeof exp === "undefined" || +exp === 0) {
        return Math.round(value)
    }
    value = +value
    exp = -(+exp)
    // If the value is not a number or the exp is not an integer...
    if(isNaN(value) || !(typeof exp === "number" && exp % 1 === 0)) {
        return NaN
    }
    // Shift
    value = value.toString().split("e")
    value = Math.round(+(value[0] + "e" + (value[1] ? (+value[1] - exp) : -exp)))
    // Shift back
    value = value.toString().split("e")
    return +(value[0] + "e" + (value[1] ? (+value[1] + exp) : exp))
}

export function setVar(name, value, variables) {
    if(variables) variables[name] = value
    return value
}

export function arrayIndex(array, index) {
    return array[index | 0]
}

export function max(array) {
    if(arguments.length === 1 && Array.isArray(array)) {
        return Math.max.apply(Math, array)
    } else if(arguments.length >= 1) {
        return Math.max.apply(Math, arguments)
    } else {
        throw new EvalError(qsTranslate("error", "Function %1 must have at least one argument.").arg("max"))
    }
}

export function min(array) {
    if(arguments.length === 1 && Array.isArray(array)) {
        return Math.min.apply(Math, array)
    } else if(arguments.length >= 1) {
        return Math.min.apply(Math, arguments)
    } else {
        throw new EvalError(qsTranslate("error", "Function %1 must have at least one argument.").arg("min"))
    }
}

export function arrayMap(f, a) {
    if(typeof f !== "function") {
        throw new EvalError(qsTranslate("error", "First argument to map is not a function."))
    }
    if(!Array.isArray(a)) {
        throw new EvalError(qsTranslate("error", "Second argument to map is not an array."))
    }
    return a.map(function(x, i) {
        return f(x, i)
    })
}

export function arrayFold(f, init, a) {
    if(typeof f !== "function") {
        throw new EvalError(qsTranslate("error", "First argument to fold is not a function."))
    }
    if(!Array.isArray(a)) {
        throw new EvalError(qsTranslate("error", "Second argument to fold is not an array."))
    }
    return a.reduce(function(acc, x, i) {
        return f(acc, x, i)
    }, init)
}

export function arrayFilter(f, a) {
    if(typeof f !== "function") {
        throw new EvalError(qsTranslate("error", "First argument to filter is not a function."))
    }
    if(!Array.isArray(a)) {
        throw new EvalError(qsTranslate("error", "Second argument to filter is not an array."))
    }
    return a.filter(function(x, i) {
        return f(x, i)
    })
}

export function stringOrArrayIndexOf(target, s) {
    if(!(Array.isArray(s) || typeof s === "string")) {
        throw new Error(qsTranslate("error", "Second argument to indexOf is not a string or array."))
    }

    return s.indexOf(target)
}

export function arrayJoin(sep, a) {
    if(!Array.isArray(a)) {
        throw new Error(qsTranslate("error", "Second argument to join is not an array."))
    }

    return a.join(sep)
}

export function sign(x) {
    return ((x > 0) - (x < 0)) || +x
}

const ONE_THIRD = 1 / 3

export function cbrt(x) {
    return x < 0 ? -Math.pow(-x, ONE_THIRD) : Math.pow(x, ONE_THIRD)
}

export function expm1(x) {
    return Math.exp(x) - 1
}

export function log1p(x) {
    return Math.log(1 + x)
}

export function log2(x) {
    return Math.log(x) / Math.LN2
}